import re

from django.conf import settings
from django.db import connections, DEFAULT_DB_ALIAS
from django.db.migrations.recorder import MigrationRecorder


MIGRATION_MODULE_TEMPLATE = '[\w\-\.]*/{}/[\w\-\. ]+\.py$'


def filter_migration_files(files_list):
    """
    Separate migration files from a list of files.

    Args:
    files_list (list): a list of file names.

    Returns:
    list: files corresponding to migration files.
    """
    # Remove pycache files from list
    clean_files_list = list(
        filter(lambda f: not re.compile('.*__pycache__.*').match(f), files_list)
    )

    # Get list of migration folders
    default_migration_folder = MIGRATION_MODULE_TEMPLATE.format('migrations')
    other_migration_folders = [
        MIGRATION_MODULE_TEMPLATE.format(module.replace('.', '/'))
        for _, module in settings.MIGRATION_MODULES.items()
    ]

    # Compile regex that matches migration folders
    generic_migration_folder_list = other_migration_folders + [default_migration_folder]
    generic_migration_folder_regex = re.compile('|'.join(generic_migration_folder_list))

    return list(filter(lambda f: generic_migration_folder_regex.match(f), clean_files_list))


def extract_migration_info_from_path(path):
    """
    Extract the migration app and name from a migration path, with format:
        [.../]<app>/<migrations-module>/<migration-name>.py

    Args:
    path (string): migration file path.

    Returns:
    app, name (tuple): migration information
    """
    path_split = path.split('/')
    app = path_split[-3]
    name = path_split[-1].replace('.py', '')
    return app, name


def is_migration_applied(app, name):
    """
    Check on migrations database table if the given migration is applied to the database.

    Args:
    app (string): django app to which the migration belongs.
    name (string): migration name.

    Returns:
    bool: True, if the migration is applied.
    """
    db_connection = connections[DEFAULT_DB_ALIAS]
    recorder = MigrationRecorder(db_connection)
    migrations = recorder.applied_migrations()
    return migrations.intersection({(app, name)})


def get_migration_operations(file_content, app, name):
    """
    Execute file content (expecting it is a migration file), instantiate a migraiton and get its
    operations attribute.

    Args:
    file_content (string): a whole migration file content.
    app (string): name of app to which the migration belongs.
    name (string): migration name.

    Returns:
    list: migration operations for given file content, app and name.
    """
    exec(file_content, globals())

    if globals().get('Migration'):
        migration = Migration(app, name)
        return migration.operations

    return []



STYLES = {
    'RED': '\033[1;31m',
    'BLUE': '\033[1;34m',
    'GREEN': '\033[0;32m',
    'RESET': '\033[0;0m',
    'BOLD': '\033[;1m',
}


def print_color(text, color):
    color_code = STYLES.get(color)
    print(color_code, end='')
    print(text)
    print(STYLES['RESET'], end='')


import os
import sys

from django.core.management.base import BaseCommand

from django_nomad.git.exceptions import GitHookAlreadyExists
from django_nomad.git.utils import find_git_directory


class Command(BaseCommand):
    help = """
        Creates a post-checkout executable file inside git hooks folder that checks migration
        files.
    """

    git_path = find_git_directory()
    hooks_path = os.path.join(git_path, 'hooks')
    post_checkout_path = os.path.abspath(os.path.join(hooks_path, 'post-checkout'))

    def add_arguments(self, *args):
        pass

    def handle(self, *args, **kwargs):
        if self.has_post_checkout_file():
            raise GitHookAlreadyExists()

        self.copy_hook_to_post_checkout_folder()
        print('post-checkout file was created...')

    def has_post_checkout_file(self):
        """
        Verify if there is a file called post-checkout inside the git hooks folder

        Returns:
        bool: True, if the file is exists.
        """
        return os.path.exists(self.post_checkout_path)

    def create_user_env_python_shebang(self):
        """
        Generate a python shebang string based on user environment (if using virtualenv for
        instance). Source copied from `flake8/src/flake8/main/git.py`.

        Returns:
        string: the formatted shebang
        """
        if sys.executable is not None:
            return sys.executable
        return '/usr/bin/env python'

    def copy_hook_to_post_checkout_folder(self):
        """
        Create a post-checkout file and add content from `post-checkout.py` to it, adding the a
        shebang based on user environment to top of file.
        """
        with open(self.post_checkout_path, 'w') as f:
            shebang = self.create_user_env_python_shebang()
            f.write(HOOK_TEMPLATE.format(shebang=shebang))
        os.chmod(self.post_checkout_path, 0o555)


HOOK_TEMPLATE = """#!{shebang}

import os
import sys

if __name__ == '__main__':
    if len(sys.argv) > 3:
        current = sys.argv[1]
        target = sys.argv[2]
        is_branch = int(sys.argv[3])

        if not is_branch:
            sys.exit(0)
        else:
            output = os.system('python manage.py check_nomad_migrations %s %s' % (current, target))
            if output > 0:
                print('An error happened checking migrations.')
"""



from django.core.management.base import BaseCommand

from django_nomad.git.utils import common_ancestor, diff_files, get_file_content_from_commit
from django_nomad.migration.utils import (
    extract_migration_info_from_path,
    filter_migration_files,
    get_migration_operations,
    is_migration_applied,
)
from ..utils import print_color


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('current')
        parser.add_argument('target')

    def handle(self, *args, **kwargs):
        print_color('Verifying migrations...', 'BLUE')

        current = kwargs['current']
        target = kwargs['target']

        # Find all migration files between previous branch and common ancestor commit
        ancestor_commit = common_ancestor(current, target)
        files_diff = diff_files(ancestor_commit, current)
        migration_files = filter_migration_files(files_diff)

        # Filter applied migrations only
        applied_migration_files = []
        for migration_file in migration_files:
            app, name = extract_migration_info_from_path(migration_file)
            if is_migration_applied(app, name):
                applied_migration_files.append((migration_file, app, name))

        # Get migration operations and feedback print messages
        if len(applied_migration_files) > 0:
            print_color('There are nomad migrations applied to this branch:', 'RED')

            for migration in applied_migration_files:
                print_color('    {}'.format(migration[0]), 'RED')
                print_color('    Operations:'.format(migration[0]), 'RED')

                file_content = get_file_content_from_commit(migration[0], current)
                operations = get_migration_operations(file_content, migration[1], migration[2])
                for operation in operations:
                    print_color('        {}'.format(operation), 'RED')

                print('')
        else:
            print_color('No nomad migrations found', 'GREEN')



import unittest

from .utils import common_ancestor, diff_files
from .exceptions import GitException


class TestGitModule(unittest.TestCase):
    # Tests assume there is a branch named 'master' and no branch named 'lalalalalalala'.

    def test_common_ancestor(self):
        # Test that default argument values don't break and return the expected type.
        self.assertIsInstance(common_ancestor('master'), str)

        # Test that if an inexistent branch is passed as argument, it raises an error.
        with self.assertRaises(GitException):
            common_ancestor('master', 'lalalalalalala')

        # Test that when two valid branches with an ancestor are passed as arguments, it returns
        # the expected type.
        self.assertIsInstance(common_ancestor('master', 'master'), str)

    def test_diff_files(self):
        # Test that when two valid branches are passed as arguments, it returns the expected type.
        self.assertTrue(diff_files('master', 'master').__iter__)

        # Test that if an inexistent branch is passed as argument, it raises an error.
        with self.assertRaises(GitException):
            common_ancestor('master', 'lalalalalalala')



import subprocess

from .exceptions import GitDirNotFound, GitException


def common_ancestor(target, current='HEAD'):
    """
    Find the most recent ancestor commit that is shared between the two branches. This function
    simply calls `git-merge-base` command.

    Args:
    target (string): name of branch to compare to current.
    current (string): name of current branch. Defaults to HEAD.

    Returns:
    string: the ancestor commit SHA-1, removing the final blank-line character.

    Raises:
    GitException: if git-merge cannot find a common ancestor.
    """
    try:
        output = subprocess.check_output(
            ['git', 'merge-base', current, target],
            stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        raise GitException(e.output.decode('utf-8')[:-1])
    else:
        return output.decode('utf-8')[:-1]


def diff_files(target, current='HEAD'):
    """
    Get list of changed files between two commit refs.

    Args:
    target (string): name of branch to compare to current.
    current (string): name of current branch. Defaults to HEAD.

    Returns:
    list: name of files that were changed between current and target.

    Raises:
    GitException: if any error occur while executing diff.
    """
    try:
        bin_output = subprocess.check_output(
            ['git', 'diff', current, target, '--name-only'],
            stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        raise GitException('Error getting diff between commits {} and {}'.format(
            current,
            target
        ))
    else:
        output = bin_output.decode('utf-8')

        # Remove empty strings
        return list(filter(bool, output.split('\n')))


def get_file_content_from_commit(file_name, commit_ref):
    """
    Get the content a file from a given commit reference.

    Args:
    file_name (string): the file path.
    commit_ref (string): the commit SHA-1 reference.

    Returns:
    string: the given file content.

    Raises:
    GitException: if any error occur while executing show.
    """
    try:
        bin_output = subprocess.check_output(
            ['git', 'show', '{}:{}'.format(commit_ref, file_name)],
            stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        raise GitException('Could not get file {} from {}'.format(file_name, commit_ref))
    else:
        return bin_output.decode('utf-8')


def find_git_directory():
    """
    Search for git directory (in case the user in not on the project root).

    Returns:
    string: path to git directory
    """
    try:
        bin_output = subprocess.check_output(
            ['git', 'rev-parse', '--git-dir'],
            stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        raise GitDirNotFound()
    else:
        return bin_output.decode('utf-8')[:-1]



class GitException(Exception):
    """
    Git-command-related exception.
    """
    pass


class GitDirNotFound(GitException):
    """
    A git folder was not found, probably the user haven't initialized git repo.
    """
    message = 'Cannot find git directory, have you run: `git init`?'


class GitHookAlreadyExists(GitException):
    """
    A hook file with given name already exists.
    """
    message = 'A post-checkout file already exists, remove it and run command again'



from setuptools import setup, find_packages


setup(
    author='Bernardo Smaniotto',
    author_email='bernardo@cheesecakelabs.com',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities'
    ],
    description='Easily handle migrations in dev mode',
    url='https://github.com/CheesecakeLabs/django-nomad',
    include_package_data=True,
    install_requires=['django'],
    name='django-nomad',
    packages=find_packages(),
    version='0.0.3'
)
